//Source file: D:/Projects/Isolde/parser/Parser.java

package parser.FSParser;

import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Hashtable;
import domainModel.*;

/**
The parser can run any finite-state grammar similar to the ones shown in the current 
grammar files.  We chose a finite-state mechanism because it is simple and sufficient
for the task.  We will need grammars for the simple default case (3-word expressions -
agent-process-patient) as well as an action grammar (see action.grammar) and a state
grammar.

Notes: 
   - The code assumes that the "process" constituent is the verb and that all
     others are case-roll fillers of some sort.

   - The parser has to know about the guards on the transitions - these can't be 
     declarative.  We'll assume noun/verb/preposition/word.

   - The parser needs an accessor to return the constituents table numbers to the GUI 
     interface so it can set the heads and constituent boundaries properly.
*/
public class Parser
{
  private Grammar _grammar;          // the finite-state grammar we are currently using
  public DmManager _dmManager;       // pointer to the domain model
  private Vector _constituents;      // list of constituents found in the grammar
  private Hashtable name2Index;      // maps from constituent name to constituent table index
  private String _cnlString;         // the string being parsed

  public Parser(DmManager inDmManager, String filename)
  {
    // Set up the grammar - parse() will initialize the grammar.
    _grammar = new Grammar(filename);
    _dmManager = inDmManager;

    // Set up the constituent table and map hash - parse() will initialize the indexes.
    _constituents = new Vector();
    name2Index = new Hashtable();
    Vector constituentNames = new Vector();
    constituentNames = _grammar.constituents();
    for (int i = 0; i < constituentNames.size(); i++)
      {
	Constituent tempConstituent =
	  new Constituent(((String)(constituentNames.elementAt(i))));
	_constituents.add(((Constituent)(tempConstituent)));
	name2Index.put(constituentNames.elementAt(i), new Integer(i));
      }

  }

  /**
    Takes a string, parses it, and creates DM concepts/instances as appropriate.
    It returns a boolean indicating if the parse worked or not.

    It will include its own view (i.e., the parser interface).

    For now, it will "reparse" by clearing the previous parse and starting from scratch.
    It will set all the constituent starts to length(parseString).
    initialize the grammar
    tokenize the input string
    set currentPosition and workingBufferStart = 0
    gets list of guards, picks one, transitions the grammar - if none works,
    do an unknown transition.
    gets list of required operations and does them.
    @roseuid 395A007A02AB
    */
  public boolean parse(String cnlString)
  {
    _cnlString = cnlString;

    //    System.out.println("parsing: " + cnlString);

    // Some working variables for the parser
    int currentPosition;         // a cnlString index of the current parser position
    int workingBufferStart;      // an index to the beginning of the current constituent
    StringTokenizer tokenizedInput; // tokenized list of input words to parse
    Vector possibleTransitions;  // a vector of outgoing links from the current state
    String currentToken;         // the current word of the input
    String currentTransition;    // the chosen transition
    String nounReference;        // the name of the DM object instance found by isNoun
    String verbReference;        // the name of the DM concept instance found by isVerb
    Vector currentOperations;    // operations to perform as specified by the grammar
    String headReference;        // name of the DM thing referenced by the head
    String previousState;        // name of the previous state in the parse
    int headLength;              // number of words of the constituent head

    // Initialize the grammar, the constituent table, the context variables
    //     and the tokenized input.
    _grammar.initialize();
    for (int i = 0; i < _constituents.size(); i++)
      ((Constituent)(_constituents.elementAt(i))).initialize();
    currentPosition = 0;
    workingBufferStart = 0;
    if (cnlString.equals(""))
      return _grammar.inTerminalState();
    tokenizedInput = new StringTokenizer(cnlString);

    // Loop through the words of the input to drive the grammar.
    while (tokenizedInput.hasMoreTokens()) {
      currentToken = tokenizedInput.nextToken();
      possibleTransitions = _grammar.possibleTransitions();
      currentTransition = "unknown";   // default
      headReference = "";              // default
      headLength = 0;                  // default
      //System.out.println("   current token: |" + currentToken + "|");

      // determine the POS of the input (there could be more than one of them I suppose)
      nounReference = isNoun(currentToken, tokenizedInput);
      verbReference = isVerb(currentToken);

      // pick a transition that matches the POS and take it
      for (int i = 0; i < possibleTransitions.size(); i++) {
	// Check for a noun match (agent or object)
	if ((possibleTransitions.elementAt(i).equals("noun")) &&
	    (nounReference != "")) {
	  currentTransition = "noun";
	  headReference = nounReference;
	  headLength = nounReference.length();
	  break;
	}
	// Check for a verb/process match
	else if ((possibleTransitions.elementAt(i).equals("verb")) &&
		 (verbReference != "")) {
	  currentTransition = "verb";
	  headReference = verbReference;
	  headLength = verbReference.length();
	  break;
	}
	// Check for a literal word match (e.g., "at", "from")
	else if (literalTransitionMatch(currentToken,
					((String)(possibleTransitions.elementAt(i))))) {
	  currentTransition = ((String)(possibleTransitions.elementAt(i)));
	  break;
	}
	// Check for simple word guards (for simple grammars that don't care about POS)
	else if (possibleTransitions.elementAt(i).equals("word")) {
	  currentTransition = "word";
	  break;
	}
      }

      //System.out.println("      perform transition: " + currentTransition);
      previousState = _grammar.getCurrentState();  // save this for one of the operations
      currentOperations = _grammar.transition(currentTransition);

      // perform the operations specified by the grammar
      for (int i = 0; i < currentOperations.size(); i++)
	switch (((Integer)(currentOperations.elementAt(i))).intValue())
	  {
	  case Transition.INCREMENT_CURRENT:
	    currentPosition += currentToken.length() + 1;
	    break;
	  case Transition.START_EQ_CURRENT:
	    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(_grammar.getCurrentState()))).intValue()))).setStartPosition(currentPosition);
	    break;
	  case Transition.START_EQ_WORKING:
	    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(_grammar.getCurrentState()))).intValue()))).setStartPosition(workingBufferStart);
	    break;
	  case Transition.WORKING_EQ_START:
	    workingBufferStart = ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(_grammar.getCurrentState()))).intValue()))).startPosition();
	    break;
	  case Transition.END_EQ_CURRENT_MINUS_1:
	    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(previousState))).intValue()))).setEndPosition(currentPosition - 1);
	    break;
	  case Transition.HEAD_EQ_CURRENT:
	    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(_grammar.getCurrentState()))).intValue()))).setHeadPosition(currentPosition);
	    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(_grammar.getCurrentState()))).intValue()))).setHeadLength(headLength);
	    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(_grammar.getCurrentState()))).intValue()))).setDomainObjectName(headReference);
	    break;
	  }
    }

    //System.out.println("out of the loop in state: " + _grammar.getCurrentState());

    // fix one special case: the last sentence constituent will have a
    // startPosition and a -1 endPosition
    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(_grammar.getCurrentState()))).intValue()))).setEndPosition(cnlString.length());

    //printConstituentTable();

    // return the appropriate status
    return _grammar.inTerminalState();
  }

  /**
    Checks the list of known GUI and domain objects

    returns the name of the DM object that was recognized

    It should match the longest possible DM object (e.g., "task model" rather than
    just "task"). and it should return the length of that object name (using the
    length of the internal name is not good enough).
    @roseuid 395A021B0097
    */
  public String isNoun(String token, StringTokenizer st)
  {
    return isNoun(token);  // deal with multiple word entities later - kvlinden, 9jan2001
  }

  private String isNoun(String token)
  {
    // Check the domain model for the 3 "types" of objects.
    // We shouldn't have to do this, there should be one "object" parent concept.
    if ((_dmManager.getEConcept(token) != null)
	|| (_dmManager.getUserConcept(token) != null)
	|| (_dmManager.getSystemConcept(token) != null)
	 )
      return token;
    else
      return "";
  }


  /**
    checks the list of predefined GUI verbs.  We may also want to use a list of
    "general" verbs from wordnet or something like that.

    returns the length of the recognized string and a pointer to the concept for
    that process (if it exists).
    @roseuid 395A0243035C
    */
  public String isVerb(String token)
  {
    if (_dmManager.getAConcept(token) != null)
      return token;
    else
      return "";
  }

  /**
    returns true if the current token matches a literal transition (surrounded by quotes
    in the grammar)
    @roseuid 395A024902EC
    */
  public boolean literalTransitionMatch(String token, String transition)
  {
    return (transition.startsWith("\"") &&
	    token.equals(transition.substring(1,transition.length() - 1)));
  }

  // a utility for debugging.
  public void printConstituentTable()
  {
    for (int i = 0; i < _constituents.size(); i++)
      ((Constituent)(_constituents.elementAt(i))).print();
  }

  /**
    goes through the constituent start positions, creating DM concepts and instances
    as required.  It is careful not to "recreate" entities when they already exist in the DM.
    It is hard-coded for the constituents we know to be in the grammar.
    @roseuid 395A07B50266
    */
  public void createDM()
  {
    DmConcept dc;       // a utility variable for creating DmConcept objects
    ActionInstance ai;  // another utility variable for entity instances
    EntityInstance ei;  // another utility variable for entity instances
    String cn;          // a temporary constituent name string

    // grab the process constituent
    Constituent c = ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get("process"))).intValue())));

    // create a new process instance - there has to be one no matter what
    if (c.startPosition() == -1) { 
      // there is no process, so it must all be canned
      ai = new ActionInstance(_cnlString);
      // change this to the "generic"/"canned" action - kvlinden, 8jan2001
      ai.setConceptOf(DmConcept.getDefaultActionConcept());  // canned strings are "generic"
    }

    else if (c.headPosition() == -1) { 
      // there is a process, but no head
      ai = new ActionInstance(_cnlString.substring(c.startPosition(), 
						   c.endPosition()));
      // change this to the "generic"/"canned" action - kvlinden, 8jan2001
      ai.setConceptOf(DmConcept.getDefaultActionConcept());  // canned strings are "generic"
    }

    else { 
      // There is a process and a specified head

      // Make sure there is a domain concept for the specified head.  
      dc = _dmManager.getAConcept(c.domainObjectName());
      if (dc == null) {
	// the DM doesn't contain the required concept, make one now & put it in the DM
	dc = new DmConcept(c.domainObjectName(), ConceptType.getDefaultActionType()); //"action type");
	dc.setParent(DmConcept.getDefaultActionConcept());
	_dmManager.addConcept(dc);
      }

      // link the process instance to its action concept
     	ai = _dmManager.getAInstanceByName(_cnlString);
      if (ai == null)
      	ai = new ActionInstance(_cnlString);
      	
      ai.setConceptOf(dc);

      // link the pre and post modifiers (if there are any)
      if (c.headPosition() != c.startPosition())
	ai.setPreModifier(_cnlString.substring(c.startPosition(),
					       c.headPosition()-1));
      if ((c.headPosition() + c.headLength()) != c.endPosition())
	ai.setPostModifier(_cnlString.substring(c.startPosition(),
						c.headPosition()-1));

      // Deal with the objects (agent, patient, location, etc.)
      //    These may come up with canned or non-canned processes, 
      //    so we do this outside of the loop.
      for (int i = 0; i < _constituents.size(); i++) {
			cn = ((Constituent)(_constituents.elementAt(i))).name();
			if (!cn.equals("process"))
			ai.setActionRelation(cn, createDMObject(cn));
      }
    }
    // now that it is fixed up, add the action instance.
    _dmManager.addAInstance(ai);
    return;
  }

  // This will return a new or existing object for the DM.
  private EntityInstance createDMObject(String constituentName)
  {
    DmConcept dc;              // a utility variable for creating DmConcept objects
    EntityInstance ei = null;  // the entity being created/referenced

    // grab the constituent so that it's easy to refer to
    Constituent c = ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(constituentName))).intValue())));

    // skip non-existent constituents
    if (c.startPosition() != -1) {

      if (c.headPosition() == -1) {
	// There is a constituent, but no specified head, so create a canned instance.
	ei = new EntityInstance(_cnlString.substring(c.startPosition(), 
						     c.endPosition()));
	ei.setConceptOf(DmConcept.getDefaultEntityConcept());  // canned entities are "thing"s
      }
      else {
	// There is a constituent with a specified head

	// If an instance with the same name and same head exists, return it.
	// We check the full instance name rather than the c.domainObjectName()
	// because action instances use the full name tp distinguish them.
	ei = _dmManager.getEInstanceByName(_cnlString.substring(c.startPosition(), 
								c.endPosition()));
	if ((ei != null) && 
	    ei.getConceptOf().getName().equals(c.domainObjectName()))
	  return ei;
	// Otherwise, make a new instance.

	// First, make sure a concept exists for the head.
	dc = _dmManager.getEConcept(c.domainObjectName());
	if (dc == null) {
	  dc = new DmConcept(c.domainObjectName(), ConceptType.getDefaultEntityType()); //"entity type");
	  dc.setParent(DmConcept.getDefaultEntityConcept());
	  _dmManager.addConcept(dc);
	}

	// Next, make the instance and connect it to the appropriate concept.
	ei = new EntityInstance(_cnlString.substring(c.startPosition(), 
						     c.endPosition()));
	ei.setConceptOf(dc);

	// Finally, add pre/post modifiers if necessary.
	if (c.headPosition() != c.startPosition())
	  ei.setPreModifier(_cnlString.substring(c.startPosition(), 
						 c.headPosition()-1));
	if ((c.headPosition() + c.headLength()) != c.endPosition())
	  ei.setPostModifier(_cnlString.substring(c.headPosition() +
						  c.headLength() + 1,
						  c.endPosition()));
      }
    }
    // Add the entity instance to the DM, the calling program will link 
    // it to its action instance.
    _dmManager.addEInstance(ei);
    return ei;
  }

  /**
    This function will assume that it will never be asked to move a start forward or
    backward beyond the next start in the default consituent order.

    The constituent will either be agent/process/patient/prepositional phrase.
    @roseuid 395B42DC03C8
    */
  public boolean moveConstituent(String constituentName, Integer newStartPosition, Integer newEndPosition)
  {
    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(constituentName))).intValue()))).setStartPosition(newStartPosition.intValue());
    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(constituentName))).intValue()))).setEndPosition(newEndPosition.intValue());
    return false;
  }

  /**
    Changes the head of the passed constituent.
    @roseuid 395B49E3004A
    */
  public boolean changeHead(String constituentName, Integer newHeadStart, Integer newHeadEnd)
  {
    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(constituentName))).intValue()))).setHeadPosition(newHeadStart.intValue());
    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(constituentName))).intValue()))).setHeadLength(newHeadEnd.intValue() - newHeadStart.intValue());
    ((Constituent)(_constituents.elementAt(((Integer)(name2Index.get(constituentName))).intValue()))).setDomainObjectName(_cnlString.substring(newHeadStart.intValue(),
																	       newHeadEnd.intValue()));
    return false;
  }

  public String cnlString()
  {
    return _cnlString;
  }

  public Vector constituents()
  {
    return _constituents;
  }

  public Vector possibleTransitions()
  {
    return _grammar.possibleTransitions();
  }
}
